@@ -8,9 +8,10 @@
 #include <qpdf/QPDFPageDocumentHelper.hh>
 #include <qpdf/QPDFWriter.hh>
 #include <qpdf/QUtil.hh>
-#include <string>
+
 #include <iostream>
-#include <cstdlib>
+#include <stdlib.h>
+#include <string>
 
 static char const* whoami = 0;
 static bool static_id = false;
@@ -77,7 +78,7 @@ int main(int argc, char* argv[])
     {
         process(whoami, argv[1], argv[2]);
     }
-    catch (std::exception e)
+    catch (std::exception const& e)
     {
         std::cerr << whoami << ": exception: " << e.what() << std::endl;
         return 2;
